# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

# various regularly used functions 

# linear interpolation at x between two points (x0,y0) and (x1,y1)
# y=(y1-y0)/(x1-x0)*(x-x0)+y0
# Result is always a float
proc interpolate {x0 y0 x1 y1 x} {
  return [expr {double($y1-$y0)/($x1-$x0)*($x-$x0)+$y0}]
}

# calculate the step variable for linear or logarithmic spacing
# Returns not a value, but an "expression", like {+2} or {*1.1}
# The functions make sure that the result is double.
# Usage: 
# set N 100
# set f0 20
# set f1 20e3
# set k [getStep $f0 $f1 $N log]
# for {set i 0; set f $f0} {$i < $N} {incr i; set f [expr $f $k]} {
#   ...
# }
#
#
proc getStep {from to n spacing} {
  if {$spacing eq {lin}} {
    set step "+ [expr {double($to - $from)/($n-1)}]"
  } elseif {$spacing eq {log}} {
    set step "* [expr {pow(10, log10(double($to)/$from)/($n-1))}]"
  } else {
    return -code error "Unknown spacing \"$spacing\""
  }
  return $step
}

# Calculates a logarithmically spaced frequency response from a fft'ed esweep impulse response or a linear spaced frequency response. Microphone compensation can also be done here. 
# Linear interpolation (maybe we should use a better one)
# Usage: 
# ...
# esweep::fft -inplace fr
# set f0 20
# set f1 20e3
# set N 200
# set logfr [getLogSpacedFr $fr [list $f0 $f1 $N]]
# To have subsequent frequency responses with identical spacing you can use this log spaced frequency response
# as $param: 
# esweep::fft -inplace fr2
# set logfr2 [getLogSpacedFr $fr2 $logfr]
proc getLogSpacedFr {fr {param} {comp {}}} {

  # interpolate frequency response
  if {[llength $param] == 3} {
    lassign $param f0 f1 N
    set k [getStep $f0 $f1 $N log]
    if {[catch {set L [esweep::size -obj $fr]}]} {
      # the supplied frequency response is not an esweep object, assume a list of format
      # frequency amplitude phase frequency ...
      # with linear mapping
      set df [lindex $fr 3]
      for {set i 0; set f $f0} {$i < $N} {incr i; set f [expr $f $k]} {
        # the index in the FR which frequency is just below $f
        set j [expr {3*int($f/$df)}]
        lassign [lrange $fr $j $j+2] w0 a0 p0
        incr j 3
        lassign [lrange $fr $j $j+2] w1 a1 p1
        set a [interpolate $w0 $a0 $w1 $a1 $f]
        set p [interpolate $w0 $p0 $w1 $p1 $f]
        lappend coords $f $a $p
      }
    } else {
      set df [expr {double([esweep::samplerate -obj $fr])/$L}]
      for {set i 0; set f $f0} {$i < $N} {incr i; set f [expr $f $k]} {
        # the index in the FR object which frequency is just below $f
        set j [expr {int($f/$df)}]
        lassign [esweep::index -obj $fr -index $j] w0 a0 p0
        if {$j == 0} {set a0 0.0}
        # incrementing this index implies that now we are just above $f
        incr j
        lassign [esweep::index -obj $fr -index $j] w1 a1 p1
        set a [interpolate $w0 $a0 $w1 $a1 $f]
        set p [interpolate $w0 $p0 $w1 $p1 $f]
        lappend coords $f $a $p
      }
    }
  } elseif {[llength $param] > 3 && [llength $param] % 3 == 0} {
    set L [esweep::size -obj $fr]
    set df [expr {double([esweep::samplerate -obj $fr])/$L}]
    set N [expr {[llength $param]/3}]
    foreach {f _ _} $param {
      set j [expr {int($f/$df)}]
      lassign [esweep::index -obj $fr -index $j] w0 a0 p0
      # incrementing this index implies that now we are just above $f
      incr j
      lassign [esweep::index -obj $fr -index $j] w1 a1 p1
      set a [interpolate $w0 $a0 $w1 $a1 $f]
      set p [interpolate $w0 $p0 $w1 $p1 $f]
      lappend coords $f $a $p
    }
  } else {
    return -code error {No valid spacing parameters}
  }

  # apply microphone correction
  if {($comp ne {}) || ([llength $comp] >= $N)} {
    # microphone correction is linearly spaced
    set df [lindex $comp 3]
    for {set i 0} {$i < [llength $coords]} {incr i 3} {
      # actually this is the same shit as above
      set f [lindex $coords $i]
      set j [expr {3*int($f/$df)}]
      lassign [lrange $comp $j $j+2] w0 a0 p0
      incr j 3
      lassign [lrange $comp $j $j+2] w1 a1 p1
      set a [interpolate $w0 $a0 $w1 $a1 $f]
      set p [interpolate $w0 $p0 $w1 $p1 $f]
      lset coords $i+1 [expr {[lindex $coords $i+1]-$a}]
      lset coords $i+2 [expr {[lindex $coords $i+2]-$p}]
    }
  } 

  return $coords
}

proc calcGatedIR {ir gate window} {
  set FS [esweep::samplerate -obj $ir]
  lassign [lmap t $gate {expr {$t > 0.0 ? int(0.5+$t*$FS/1000.0) : 0}}] gate_start gate_stop
  if {$gate_stop >= [esweep::size -obj $ir]} {set gate_stop [expr {[esweep::size -obj $ir]-1}]}
  set gate_length [expr {$gate_stop-$gate_start}]
  set ir_gated [esweep::create -type [esweep::type -obj $ir] -size $gate_length -samplerate $FS]
  esweep::copy -dst ir_gated -src $ir -srcIdx $gate_start
  esweep::window -obj ir_gated \
    -rightwin [lindex $window 0] -rightwidth [lindex $window 1]

  return $ir_gated
}

proc calcFFT {ir sc size smoothing} {
  set FS [esweep::samplerate -obj $ir]

  set gate_length [esweep::size -obj $ir]
  if {$size > 0} {
    if {$gate_length > $size} {
      return -code error "Gate length $gate_length exceeds FFT size $size"
    } else {
      set fft_size $size
    }
  } else {
    set fft_size 0
    while {[set fft_size [expr {$fft_size > 0 ? $fft_size << 1 : 1}]] < $gate_length} {continue}
  }
  set fr [esweep::create -type [esweep::type -obj $ir] -size $fft_size -samplerate $FS]
  # copy instead of clone, this automatically zero-pads the impulse response
  esweep::copy -dst fr -src $ir
  esweep::expr fr * $sc
  esweep::fft -inplace -obj fr
  esweep::toPolar -obj fr
  if {$smoothing > 0} {
    esweep::smooth -obj fr -factor $smoothing
  }
  esweep::lg -obj fr
  esweep::expr fr * 20

  return $fr
}

proc calcHD {ir sc gate fft_size sweeprate window hd_list args} {
  set FS [esweep::samplerate -obj $ir]
  lassign [lmap t $gate {expr {$t > 0.0 ? int(0.5+$t*$FS/1000.0) : 0}}] gate_start gate_stop
  set gate_length [expr {$gate_stop-$gate_start}]
  if {$fft_size > 0} {
    if {$gate_length > $fft_size} {
      return -code error "Gate length $gate_length exceeds FFT size $fft_size"
    } 
  } else {
    while {[set fft_size [expr {$fft_size > 0 ? $fft_size << 1 : 1}]] < $gate_length} {continue}
  }
  # parse options
  set weighting none
  set shifting 0
  set smoothing 0
  set mode uncompressed
  foreach {opt val} $args {
    switch -- $opt {
      -weighting {
        set weighting $val
      }
      -shift {
        set shifting $val
      }
      -smoothing {
        set smoothing $val
      }
      -mode {
        set mode $val
      }
      default {
      }
    }
  }

  foreach k $hd_list {
    if {$k < 2} {continue}
    # get the position of the k'th HD IR
    set shift [expr {int(0.5+log($k)*$sweeprate*$FS)}]
    set start [expr {$gate_start-$shift}]
    # wrap-around
    set start [expr {$start < 0 ? [esweep::size -obj $ir] + $start : $start}]

    # avoid mixing with previous HD IR
    set max_length [expr {$shift-int(0.5+log($k-1)*$sweeprate*$FS)}]
    set length [expr {$gate_length > $max_length ? $max_length : $gate_length}]
    # cut out harmonic impulse response
    # wrap around at end of total IR
    set stop [expr {$start+$length}]
    if {$stop >= [esweep::size -obj $ir]} {
      set ir_gated [esweep::create -type [esweep::type -obj $ir] -size $length -samplerate $FS]
      set l [esweep::copy -src $ir -dst ir_gated -srcIdx $start]
      # ir_gated may already be filled up, so catch this 2nd copy
      catch {esweep::copy -src $ir -dst ir_gated -srcIdx 0 -dstIdx $l}
    } else {
      set ir_gated [esweep::get -obj $ir -from $start -to $stop]
    }
    esweep::window -obj ir_gated \
      -rightwin [lindex $window 0] -rightwidth [lindex $window 1]

    set fr [esweep::create -type [esweep::type -obj $ir_gated] -size $fft_size -samplerate $FS]
    esweep::copy -dst fr -src $ir_gated

          
    # weighting
    switch -- $weighting {
      linear {set scale [expr {$sc*$k/2.0}]}
      quadratic {set scale [expr {$sc*($k**2)/4.0}]}
      A -
      B -
      C -
      D -
      ITU468 {
        lassign [weighting $FS $weighting] norm filter
        esweep::filter -signal fr -filter $filter
        set scale [expr {$sc*$norm}]
      }
      none {set scale $sc}
      default {
        set msg "Weighting function $weighting unknown or not implemented. Using \"none\""
        ::cmdbar::writeToCmdBar $msg -bg yellow
        set scale $sc
      }
    }
    set scale [expr {$scale*pow(10,$shifting/20.0)}]
    esweep::expr fr * $scale

    # make frequency response
    esweep::fft -inplace -obj fr
    esweep::toPolar -obj fr
    if {$smoothing > 0} {
      esweep::smooth -obj fr -factor $smoothing
    }
    esweep::lg -obj fr
    esweep::expr fr * 20

    if {$mode eq {compressed}} {esweep::compress -obj fr -factor $k}

    lappend result $k $fr
  }
  return $result
}


proc c2p {a} {
  foreach {f re im} $a {
    lappend b $f [expr {sqrt($re**2+$im**2)}] [expr {atan2($im, $re)}]
  }
  return $b
}

proc p2c {a} {
  foreach {f m p} {
    lappend b $f [expr {$m*cos($p)}] [expr {$m*sin($p)}]
  }
  return $b
}


##############################################################
# these functions help to implement a coroutine based logger #
##############################################################
proc spawn cmd {
  set k [gensym]
  coroutine $k {*}$cmd
  return $k
}

proc gensym {{prefix "::coroutine"}} {
  variable gensymid
  return $prefix[incr gensymid]
}

proc uniqueName {{prefix ::var}} {
  # make sure that the clock is ticking away...
  after 1
  return $prefix[clock microseconds]
}
 
proc updateTimer {timerVar waiterVar {delay 100}} {
  upvar $timerVar timer
  upvar $waiterVar waiter
  catch {after cancel $timer}
  set timer [after $delay [list apply {{coro waiterVar} {
     upvar $waiterVar waiter
     set waiter 0
     DEBUGPRINT $coro
     $coro 
   }} [info coroutine] $waiterVar ]]
}

